fs = require 'fs'
crypto = require 'crypto'
async = require 'async'
restify = require 'restify'
npm_package = require '../package.json'

module.exports = exports = (cb, options={}) ->
  options.env ?= process.env.NODE_ENV || 'development'
  options.port ?= 8082
  options.setupResource ?= '/'
  options.configPath ?= './config.json'
  options.encryptedConfigPath ?= './config.ct.json'
  options.encryptKeyRegex ?= /^[\x00-\x7F]{8,128}$/
  options.encryptKeyParam ?= 'secure_setup_key'
  options.silent ?= false

  switch options.env

    #
    # Production
    #
    when 'production'
      # Plaintext configuration file MUST NOT exist in production
      if fs.existsSync options.configPath
        return cb new Error "Remove config file '#{ options.configPath }' in production"

      # Ciphertext config file MUST exist
      if not fs.existsSync options.encryptedConfigPath
        return cb new Error "Encrypted config file '#{ options.encryptedConfigPath }' missing"

      # Create server
      server = restify.createServer
        name: 'secure-setup'
      server.use restify.queryParser()
      server.use restify.bodyParser()

      # Setup route 
      server.post options.setupResource, (req, res, next) ->
        # Filter key input
        if not options.encryptKeyRegex.test req.params.key
          res.send 'Invalid key'
          return cb new Error 'Received invalid key'

        decryptConfig options.encryptedConfigPath, req.params.key, (err, conf) ->
          if err
            res.send { "message": "failed with error '#{ err }'" }
            console.err err
            return cb err
          else
            res.send { "message": "success" }
            if !options.silent
              console.log 'Configuration successfully decrypted. Passing to callback.'
            cb null, conf
            next()

      # Launch server
      server.listen options.port, ->
        if !options.silent
          console.log 'Server waiting to receive key...'
          console.log 'curl --data "key=<KEY>" %s%s', server.url, options.setupResource


    #
    # Dev
    #
    when 'development'
      # Create example config file, if it doesn't exist
      if not fs.existsSync options.configPath
        fs.writeFileSync options.configPath, '{\n\t"version":"0.0.1",\n\t"foo":"bar"\n}'
        return cb new Error "Config file #{ options.configPath } doesn't exist. Creating example."

      # Create ciphertext config file
      encryptConfig options.configPath, options.encryptedConfigPath
      , options.encryptKeyParam, (err) ->
        return cb err if err

        # Load config data
        loadObj options.configPath, (err, conf) ->
          return cb err if err
          if !options.silent
            console.log "Encrypted config with key: #{ conf[options.encryptKeyParam] }"
            console.log "You MUST remove plaintext config file before running in production"
          cb null, conf


    #
    # Default
    #
    else
      return cb new Error 'Unrecognised NODE_ENV. Set env in \'options\' argument'


## Functions
randomString = (len, encoding) ->
  encoding = encoding or 'hex'
  c = crypto.randomBytes(len).toString encoding
  c.replace(/[+/]/g, '').substring 0, len

hash = (string, salt='') ->
  hasher = crypto.createHash 'sha256'
  hasher.update salt + string
  hasher.digest 'base64'

encrypt = (plain, key, algorithm='aes-256-cbc', encoding='base64') ->
  c = crypto.createCipher algorithm, key
  cipher = c.update plain, 'utf8', encoding
  cipher += c.final encoding

decrypt = (cipher, key, algorithm='aes-256-cbc', encoding='base64') ->
  c = crypto.createDecipher algorithm, key
  plain = c.update cipher, encoding, 'utf8'
  plain += c.final 'utf8'

hasKeys = (obj, keys=[]) ->
  ret = yes
  keys.forEach (key) ->
    ret = no if not (key of obj)
  ret

redactKeys = (obj, keys=[]) ->
  keys.forEach (key) ->
    delete obj[key]
  obj

loadObj = (path, cb) ->
  fs.exists path, (exists) ->
    return cb new Error "File does not exist at path '#{ path }'" if not exists
       
    fs.readFile path, 'utf8', (err, data) ->
      return cb new Error "Error reading file '#{ path }'" if err

      obj = JSON.parse data
      if typeof obj isnt 'object'
        return cb new Error "Could not parse json in file '#{ path }'"
      cb err, obj

saveObj = (obj, path, cb) ->
  string = JSON.stringify obj, null, '\t'
  fs.writeFile path, string, (err) ->
    return cb new Error "Could not write to file '#{ path }'" if err
    cb null

encryptConfig = (plainPath, cipherPath, encryptKeyParam, cb) ->
  async.waterfall [
    # Load plaintext config file
    (cb) ->
      loadObj plainPath, (err, conf) ->
        return cb err if err

        # Append key to plaintext config, if not set
        conf[encryptKeyParam] ?= randomString 32, 'base64'
        cb err, conf

    # Overwrite config file with new super-object
    (conf, cb) ->
      saveObj conf, plainPath, (err) ->
        cb err, conf

    # Encrypt config and save in config.ct.json
    (conf, cb) ->
      confString = JSON.stringify conf
      cipherConfString = encrypt confString, conf[encryptKeyParam]
      salt = randomString 16, 'base64'
      encrConf =
        secure_setup_version: npm_package.version
        salt: salt
        hash: hash conf[encryptKeyParam], salt
        ciphertext: cipherConfString

      saveObj encrConf, cipherPath, cb

  ], (err) ->
    cb err

decryptConfig = (cipherPath, key, cb) ->
  # Load ciphertext config file
  loadObj cipherPath, (err, encrConf) ->
    return cb err if err

    # Validate config.ct.json contents
    if not hasKeys encrConf, ['secure_setup_version', 'salt', 'hash', 'ciphertext']
      return cb new Error "Encrypted config file '#{ cipherPath }' missing required params"

    # Validate decryption key
    if encrConf.hash isnt hash key, encrConf.salt
      console.error 'Decryption key incorrect'
      return cb new Error 'Decryption key incorrect'

    # Decrypt config
    confString = decrypt encrConf.ciphertext, key
    conf = JSON.parse confString
    cb null, conf